// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibus;

/// Message types that are passed through the Omnibus.
enum _OmnibusMessageType { event, request, response }

/// Omnibus message base
///
/// An omnibus message has an [omnibusId] so it can be safely identified by as a
/// specific message type. All message data is stored in the [data] Map. The map
/// should only contain JSON compatible data. A map was choosen to allow easy
/// message filtering and easy JSON conversion (as required for integrating
/// non-Dart handlers).
abstract class _OmnibusMessage {
  /// Message identifier key
  ///
  /// Use the following convention for this identifier:
  /// `package_name.MessageName`
  final String omnibusId;

  /// Message type
  final _OmnibusMessageType type;

  /// Message data
  final Map<String, dynamic> data;

  /// Default constuctor
  _OmnibusMessage(this.omnibusId, this.type, Map<String, dynamic> data)
      : data = _getJsonMap(data) {}

  /// Constructor to bypass _getPrimitiveMap, use with care!
  _OmnibusMessage.unchecked(this.omnibusId, this.type, this.data) {}
}

/// Event message
///
/// An event message is published when the module wants to notify other handlers
/// of what is going on. An example of this is a progress update. This message
/// can be used for publish-subscribe messaging.
class OmnibusEvent extends _OmnibusMessage {
  /// Default constructor
  OmnibusEvent(String omnibusId, Map<String, dynamic> data)
      : super(omnibusId, _OmnibusMessageType.event, data);

  /// Constructor to bypass [_OmnibusMessage._getPrimitiveMap]
  OmnibusEvent.unchecked(String omnibusId, Map<String, dynamic> data)
      : super.unchecked(omnibusId, _OmnibusMessageType.event, data);
}

/// Request message
///
/// A request message is published when the module requires certain data. This
/// message can be used for request-reply messasing.
class OmnibusRequest extends _OmnibusMessage {
  /// Default constructor
  OmnibusRequest(String omnibusId, Map<String, dynamic> data)
      : super(omnibusId, _OmnibusMessageType.request, data);

  /// Constructor to bypass [_OmnibusMessage._getPrimitiveMap]
  OmnibusRequest.unchecked(String omnibusId, Map<String, dynamic> data)
      : super.unchecked(omnibusId, _OmnibusMessageType.request, data);
}

/// Response message
///
/// The response message is returned by handlers in order to reply to a request
/// message.
class OmnibusResponse extends _OmnibusMessage {
  /// Default constructor
  OmnibusResponse(String omnibusId, Map<String, dynamic> data)
      : super(omnibusId, _OmnibusMessageType.response, data);

  /// Constructor to bypass [_OmnibusMessage._getPrimitiveMap]
  OmnibusResponse.unchecked(String omnibusId, Map<String, dynamic> data)
      : super.unchecked(omnibusId, _OmnibusMessageType.response, data);
}
